@font-face {
    font-family: 'Comfortaa';
    src: url("../fonts/Comfortaa-Regular.ttf") format("truetype");
}

.md-header-nav__topic .md-ellipsis {
    font-family: 'Comfortaa', cursive;
    font-weight: bold;
}

.md-nav__title--site {
    font-family: 'Comfortaa', cursive;
    font-weight: bold;
}